import json
import pandas as pd
import urllib.parse #input data
from tkinter import Tk
from tkinter import filedialog
from enum import Enum

def get_file_path_excel():
    root = Tk()
    root.update()
    def open_file():
        file = filedialog.askopenfilename(filetypes=[("excel files", "*.xlsx")]) 
        return file
    file_path = open_file()
    root.destroy()
    return file_path


def read_data_frame_from_local_excel_file():
    try:
        from google.colab import files
        IN_COLAB = True
    except:
        IN_COLAB = False
        
    if IN_COLAB:
        
        uploaded = files.upload()
        excel_file_name = list(uploaded.keys())[0]
        df = pd.read_excel(io.BytesIO(uploaded[excel_file_name]))
        return(df)
    else:
        excel_path = get_file_path_excel()
        df = pd.read_excel(excel_path)
        return(df)
    




import tensorflow as tf
import numpy as np
import os
import cv2
import matplotlib.pyplot as plt
import random
from sklearn.model_selection import train_test_split

training_data =[]

#This function separates both features & labels from training data
def seggregate_data(training_data):
    X = []
    Y = []
    for features, lables in training_data:
        X.append(features)
        Y.append(lables)
    X = np.array(X)
    Y = np.array(Y)
    return X,Y

#This function reads each folder and each image and converts into array
#This function internally calls seggregate_data to separate features & labels
def create_training_data(CATEGORIES,DATADIR, IMG_SIZE = 100):
    for category in CATEGORIES:
        class_num = CATEGORIES.index(category)
        path = os.path.join(DATADIR,category)
        for img in os.listdir(path):
            try:
                img_array = cv2.imread(os.path.join(path,img), cv2.IMREAD_GRAYSCALE) 
                new_array = cv2.resize(img_array,(IMG_SIZE,IMG_SIZE))
                training_data.append([new_array,class_num])
            except Exception as e:
                pass
    random.shuffle(training_data)
    features,labels = seggregate_data(training_data)
    return features,labels

#This function splits the data into test and train
#This function normalizes the features values in both train and test
#This function converts label values to binary matrix
def split_normalize_data(X, Y, category_count, TEST_SIZE = 0.20,IMG_SIZE=100 ):
    x_train, x_test, y_train, y_test = train_test_split(X, Y, test_size = TEST_SIZE)
    x_train = x_train.reshape(x_train.shape[0], IMG_SIZE, IMG_SIZE, 1)
    x_test = x_test.reshape(x_test.shape[0], IMG_SIZE, IMG_SIZE, 1)

    x_train = x_train.astype('float32')
    x_test = x_test.astype('float32')

    x_train/=255
    x_test/=255

    y_train = tf.keras.utils.to_categorical(y_train, category_count)
    y_test = tf.keras.utils.to_categorical(y_test, category_count)
    
    return x_train, x_test, y_train, y_test

#This function creates a model compiling input, Hidden and output layers
def image_classifier(category_count, IMG_SIZE = 100):
    
    model = tf.keras.models.Sequential()

    model.add(tf.keras.layers.Conv2D(32,(3,3),input_shape=(IMG_SIZE,IMG_SIZE,1)))
    model.add(tf.keras.layers.Dense(32,activation=tf.nn.relu))
    model.add(tf.keras.layers.MaxPooling2D(pool_size=(2,2)))

    model.add(tf.keras.layers.Conv2D(64, (3,3)))
    model.add(tf.keras.layers.Dense(64,activation=tf.nn.relu))
    model.add(tf.keras.layers.MaxPooling2D(pool_size=(2,2)))

    model.add(tf.keras.layers.Flatten())

    model.add(tf.keras.layers.Dropout(0.2))

    model.add(tf.keras.layers.Dense(category_count,activation=tf.nn.softmax))

    model.compile(optimizer='adam',
                     loss='categorical_crossentropy',
                     metrics=['accuracy'])
    return model



import logging

def experiment_design_log(design):
    logger = logging.getLogger(__name__)
    logger.setLevel(logging.INFO)
    formatter = logging.Formatter('%(asctime)s:%(name)s:%(levelname)s:%(message)s')
    file_handler = logging.FileHandler('experiment_design.log')
    file_handler.setFormatter(formatter)
    if logger.hasHandlers():
        pass
    else:
        logger.addHandler(file_handler)
    logger.info(design)


from numpy import unique
from numpy import where
from math import sqrt
from sklearn.datasets import make_classification
from sklearn.cluster import AffinityPropagation
from matplotlib import pyplot
from sklearn.metrics import silhouette_samples, silhouette_score,davies_bouldin_score
from math import sqrt
from sklearn.cluster import KMeans
from sklearn.cluster import DBSCAN

def Clustering(X,num_clusters=None):
    AP = AffinityPropagation(damping=0.9)
    AP.fit(X)
    # assign a cluster to each example
    yhat_AP = AP.predict(X)
    silhouette_AP = silhouette_score(X, yhat_AP)
    
    def optimal_number_of_clusters(wcss):
        x1, y1 = 2, wcss[0]
        x2, y2 = 20, wcss[len(wcss)-1]

        distances = []
        for i in range(len(wcss)):
            x0 = i+2
            y0 = wcss[i]
            numerator = abs((y2-y1)*x0 - (x2-x1)*y0 + x2*y1 - y2*x1)
            denominator = sqrt((y2 - y1)**2 + (x2 - x1)**2)
            distances.append(numerator/denominator)    
        return distances.index(max(distances)) + 2
    if num_clusters!=None:
        model_KM=KMeans(n_clusters=num_clusters)
        model_KM.fit(X)
        # assign a cluster to each example
        yhat_KM = model_KM.predict(X)
        silhouette_KM = silhouette_score(X, yhat_KM)
    else:
        wcss = []
        for n in range(2, 21):
            kmeans = KMeans(n_clusters=n)
            kmeans.fit(X)
            wcss.append(kmeans.inertia_)
        clusters=optimal_number_of_clusters(wcss)
        model_KM=KMeans(n_clusters=clusters)
        model_KM.fit(X)
        # assign a cluster to each example
        yhat_KM = model_KM.predict(X)
        silhouette_KM = silhouette_score(X, yhat_KM)
    
    model_DB = DBSCAN(eps=9.5, min_samples=2)
    # fit model and predict clusters
    yhat_DB = model_DB.fit_predict(X)
    silhouette_DB = silhouette_score(X, yhat_DB)
    
    silhouette=[silhouette_DB,silhouette_KM,silhouette_AP]
    algo=[model_DB,model_KM,AP]
    maxi=algo[silhouette.index(max(silhouette))]
    
    return maxi.fit(X)
    
    
        



import pandas as pd
from tpot import TPOTRegressor
from tpot import TPOTClassifier
from sklearn.model_selection import train_test_split
from feature_engine import categorical_encoders as ce
from .interpret_model import Global_Model_Explanation
from .interpret_model import Explanation_Dashboard
from sklearn import preprocessing
import warnings
import os
import numpy as np
from dxc.ai.global_variables import globals_file
from sklearn.metrics import SCORERS

###encode and return the categorical data
def categorical_encoding(data,target):
    data_1 = data.dropna()
    objFeatures = data_1.select_dtypes(include="object").columns
    objlen = len(list(objFeatures))
    if objlen == 0:             
        return data_1.drop([target], axis = 1), data_1[target]
    else:
        if data_1[target].dtype == 'object':      
            le = preprocessing.LabelEncoder()
            try:
                data_1[target] = le.fit_transform(data_1[target].astype(str))
            except:
                data_1[target] = data_1[target].astype('category')
                data_1[target] = le.fit_transform(data_1[target].astype(str))  
            globals_file.run_experiment_target_encoder = le
            globals_file.run_experiment_target_encoder_used = True
        data_2 = data_1.drop([target], axis = 1)
        encoder = ce.OrdinalCategoricalEncoder(encoding_method='ordered')
        encoder.fit(data_2, data_1[target])
        data_3 = encoder.transform(data_2)
        globals_file.run_experiment_encoder = encoder
        globals_file.run_experiment_encoder_used = True
        return data_3, data_1[target]

##Define Tpot regressor 
def regressor(verbosity, max_time_mins , max_eval_time_mins, config_dict, warm_start, scoring):
    if verbosity == True:
        model_def = TPOTRegressor(verbosity=2, max_time_mins=max_time_mins, max_eval_time_mins= max_eval_time_mins, config_dict = config_dict, warm_start = warm_start, scoring= scoring, template = 'Regressor')
    else:
        model_def = TPOTRegressor(verbosity=0, max_time_mins=max_time_mins, max_eval_time_mins= max_eval_time_mins, config_dict = config_dict, warm_start = warm_start, scoring= scoring, template = 'Regressor')
    return model_def

##Define Tpot classifier
def classifier(verbosity, max_time_mins, max_eval_time_mins, config_dict, warm_start, scoring):
    if verbosity == True:
        model_def = TPOTClassifier(verbosity=2, max_time_mins=max_time_mins, max_eval_time_mins= max_eval_time_mins, config_dict = config_dict, warm_start = warm_start, scoring = scoring, template = 'Classifier')
    else:
        model_def = TPOTClassifier(verbosity=0, max_time_mins=max_time_mins, max_eval_time_mins= max_eval_time_mins, config_dict = config_dict, warm_start = warm_start, scoring = scoring, template = 'Classifier')
    return model_def

###Train the model 
def train_model(data, target, model_def, model_type, interpret = False, warm_start = False, export_pipeline = True):
    if warm_start == False:
        globals_file.run_experiment_encoder = None
        globals_file.run_experiment_target_encoder = None
        globals_file.run_experiment_target_encoder_used = False
        globals_file.run_experiment_encoder_used = False
        globals_file.run_experiment_warm_start = False
        data_transformed, label_data = categorical_encoding(data,target)
        try:
            x_train, x_test, y_train, y_test = train_test_split(data_transformed, label_data, test_size=0.2, random_state=0, stratify = label_data)
        except:
            x_train, x_test, y_train, y_test = train_test_split(data_transformed, label_data, test_size=0.2, random_state=0) 
    ##Save the data for first time execution with warm start.
    if warm_start == True and globals_file.run_experiment_warm_start == False:
        globals_file.run_experiment_encoder = None
        globals_file.run_experiment_target_encoder = None
        globals_file.run_experiment_target_encoder_used = False
        globals_file.run_experiment_encoder_used = False
        data_transformed, label_data = categorical_encoding(data,target)
        try:
            x_train, x_test, y_train, y_test = train_test_split(data_transformed, label_data, test_size=0.2, random_state=0, stratify = label_data)
        except:
            x_train, x_test, y_train, y_test = train_test_split(data_transformed, label_data, test_size=0.2, random_state=0)
        globals_file.run_experiment_warm_start = True
        globals_file.run_experiment_x_train = x_train
        globals_file.run_experiment_x_test = x_test
        globals_file.run_experiment_y_train = y_train
        globals_file.run_experiment_y_test = y_test
        globals_file.run_experiment_datatransformed = data_transformed
        globals_file.run_experiment_labeldata = label_data
    ##Return the data from second time execution with warm start
    if warm_start == True and globals_file.run_experiment_warm_start == True:
        x_train = globals_file.run_experiment_x_train
        x_test = globals_file.run_experiment_x_test
        y_train = globals_file.run_experiment_y_train
        y_test = globals_file.run_experiment_y_test
        data_transformed = globals_file.run_experiment_datatransformed
        label_data = globals_file.run_experiment_labeldata
    
    model = model_def.fit(x_train, y_train)
    globals_file.run_experiment_model = model
    best_pipeline = model.fitted_pipeline_
    if model_type == 'TPOTRegressor':
        print()
        # r2 score
        score_r2 = SCORERS['r2'](best_pipeline, x_test, y_test)
        print("r2 Score:", score_r2)
        print()
        
        # neg_mean_squared_error
        score_nmse = SCORERS['neg_mean_squared_error'](best_pipeline, x_test, y_test)
        print('Negative mean square error:', score_nmse)
        print()
        
        # neg_root_mean_squared_error
        score_nrmse = SCORERS['neg_root_mean_squared_error'](best_pipeline, x_test, y_test)
        print('Negative root mean square error:', score_nrmse)
        print()
    
        # explained variance
        score_var = SCORERS['explained_variance'](best_pipeline, x_test, y_test)
        print('explained_variance:', score_var)
        print()
    
        # negative mean absolute error
        score_nmae = SCORERS['neg_mean_absolute_error'](best_pipeline, x_test, y_test)
        print('Negative_mean_absolute_error:', score_nmae)
        print()
        
        # Negative median absolute error
        score_nmdae = SCORERS['neg_median_absolute_error'](best_pipeline, x_test, y_test)
        print('Negative_median_absolute_error:', score_nmdae)
        print()
    
    if model_type == 'TPOTClassifier':
        print()   
        # Accuracy 
        score_acc = SCORERS['accuracy'](best_pipeline, x_test, y_test)
        print("Accuracy:", score_acc)
        print()
        
        #ROC_AUC_OVR
        try:
            score_roc_ovr = SCORERS['roc_auc_ovr'](best_pipeline, x_test, y_test)
            print("ROC_AUC_OVR:", score_roc_ovr)
            print()
        except:
            pass
        
        #ROC_AUC_OVO
        try:
            score_roc_ovo = SCORERS['roc_auc_ovo'](best_pipeline, x_test, y_test)
            print("ROC_AUC_OVO:", score_roc_ovo)
            print()
        except:
            pass

        # Recall
        score_rec_macro = SCORERS['recall_macro'](best_pipeline, x_test, y_test)
        print("Recall:", score_rec_macro)        
        print()
        
        # Precision
        score_pre_macro = SCORERS['precision_macro'](best_pipeline, x_test, y_test)
        print("Precision:", score_pre_macro)        
        print()
        
        #F1
        score_f1_macro = SCORERS['f1_macro'](best_pipeline, x_test, y_test)
        print("F1 Score:", score_f1_macro)        
        print()        
        
    
    ##Save the pipeline and data 
    if export_pipeline == True:
        currentDirectory = os.getcwd()
        currentDirectory = currentDirectory + '\data_file.csv'
        data_file_combine = data_transformed
        data_file_combine['target'] = label_data
        data_file_combine.to_csv(currentDirectory, index = False, header=True, encoding='utf-8')
        model.export('best_pipeline.py', currentDirectory)
    ##Generate interpret interactive charts
    if interpret == True:
        global_explanation = Global_Model_Explanation(model.fitted_pipeline_,x_train,x_test,feature_names = None,classes = None,                       explantion_data = None)
        Explanation_Dashboard(global_explanation, model.fitted_pipeline_, x_train, x_test, explantion_data = None)
    return model.fitted_pipeline_



import arrow #normalizing dates
import numpy as np
from sklearn.base import TransformerMixin #impute missing data
from auto_ml import Predictor #ML models
from sklearn.model_selection import train_test_split
import os
import pickle
from contextlib import redirect_stdout
import warnings
import io
from dxc.ai.global_variables import globals_file
from .TimeSeriesModels import getBestForcastingModel
from dxc.ai.logging import experiment_design_logging
from .model_pipeline import regressor
from .model_pipeline import train_model
from .model_pipeline import classifier


# define the general class of models
class model:
    __model = []
    def build(self, meta_data): raise NotImplementedError()
    def train_and_score(self, data): raise NotImplementedError()
    def interpret(self): raise NotImplementedError()
    def python_object(): raise NotImplementedError()

    @staticmethod
    def meta_data_key(meta_data, value):
        key_list = list(meta_data.keys())
        val_list = list(meta_data.values())

        return key_list[val_list.index(value)]

#define the model lifecycle


# define a prediction model
class prediction(model):

    @property
    def estimator(self):
        raise NotImplementedError()

    def build(self, meta_data, verbose, max_time_mins, max_eval_time_mins, config_dict, warm_start, scoring):
        if self.estimator == 'TPOTRegressor':
            if globals_file.run_experiment_warm_start == False or warm_start == False:
                self.__model = regressor(verbose, max_time_mins, max_eval_time_mins, config_dict, warm_start, scoring)
            else:
                self.__model = globals_file.run_experiment_model
        elif self.estimator == 'TPOTClassifier':
            if globals_file.run_experiment_warm_start == False or warm_start == False:
                self.__model = classifier(verbose, max_time_mins, max_eval_time_mins, config_dict, warm_start, scoring)
            else:
                self.__model = globals_file.run_experiment_model
        else:    
            self.__model = Predictor(type_of_estimator=self.estimator, column_descriptions=meta_data)
        self.__label = self.meta_data_key(meta_data, "output")

    def train_and_score(self, data, labels, verbose, interpret, warm_start, export_pipeline):
        ##Train and score 
        if self.estimator == 'TPOTRegressor' or self.estimator == 'TPOTClassifier':
            self.__model = train_model(data, self.__label, self.__model, self.estimator, interpret, warm_start, export_pipeline)
        else:
            # create training and test data
            training_data, test_data = train_test_split(data, test_size=0.2)

        # train the model
            if verbose == False:
                warnings.filterwarnings('ignore')
                text_trap = io.StringIO()
                with redirect_stdout(text_trap):
                    self.__model.train(training_data, verbose=False, ml_for_analytics= False)
            else:
                warnings.filterwarnings('ignore')
                self.__model.train(training_data, verbose=True, ml_for_analytics=False)

        # score the model
            if verbose == False:
                self.__model.score(test_data, test_data[self.__label], verbose=0)
            else:
                self.__model.score(test_data, test_data[self.__label], verbose=1)

    def interpret(self):
        pass

    def python_object(self):
        return self.__model

# define a regressor model
class regression(prediction):
    @property
    def estimator(self):
        return("regressor")

# define a classification model
class classification(prediction):
    @property
    def estimator(self):
        return("classifier")
    
# define a tpot classification model
class tpot_classification(prediction):
    @property
    def estimator(self):
        return("TPOTClassifier")
    
# define a Tpot regressor model
class tpot_regression(prediction):
    @property
    def estimator(self):
        return("TPOTRegressor")

def run_experiment(design, verbose = False, interpret = False, max_time_mins = 5, max_eval_time_mins = 0.04 , config_dict = None, warm_start = False, export_pipeline = True, scoring = None):
    experiment_design_logging.experiment_design_log(design)
    if design["model"] == 'timeseries':
        trained_model = getBestForcastingModel(design['labels'], no_predictions=7, debug=verbose, visualize = False)
        return trained_model
    globals_file.run_experiment_used = True
    design["model"].build(design["meta_data"], verbose, max_time_mins, max_eval_time_mins, config_dict,warm_start, scoring)
    design["model"].train_and_score(design["data"], design["labels"], verbose, interpret, warm_start, export_pipeline)
    design["model"].interpret()
    return design["model"].python_object()
    



import pandas as pd
import janitor #data cleaning
from ftfy import fix_text #data cleaning
import nltk #data cleaning
nltk.download('punkt') #data cleaning
import scrubadub #data cleaning
import arrow #normalizing dates
import numpy as np
from sklearn.base import TransformerMixin
from sklearn.impute import KNNImputer ##using KNN as imputer for categorical fields
from sklearn.preprocessing import OrdinalEncoder ##Ordinal encoder is being used for encoding categorical objects
from dxc.ai.global_variables import globals_file
from datetime import datetime

def encode(data):

    '''function to encode non-null data and replace it in the original data'''
    encoder = OrdinalEncoder()
    #retains only non-null values
    nonulls = np.array(data.dropna())
    #reshapes the data for encoding
    impute_reshape = nonulls.reshape(-1,1)
    #encode date
    impute_ordinal = encoder.fit_transform(impute_reshape)
    #encoders_store[column_name]=encoder
    #Assign back encoded values to non-null values
    data.loc[data.notnull()] = np.squeeze(impute_ordinal)
    return (data,encoder)

def impute_df(df):
    # imputer = KNN()
    imputer = KNNImputer(n_neighbors=2)
    object_types = list(df.select_dtypes(include=['object']).columns)
    num_types = list(set(df.columns) - set(object_types))
    encoders_store={}
    for column in num_types:
        skew=df[column].skew()
        if (-1 < skew < 1):
            df[column]=df[column].fillna(df[column].mean())
        else :
            df[column]=df[column].fillna(df[column].median())
    #create a for loop to iterate through each column in the data
    for columns in object_types:
        new=encode(df[columns])
        encoders_store[columns]=new[1]
    imputed_data = pd.DataFrame(np.round(imputer.fit_transform(df)),columns = df.columns)
    for columns in object_types:
        imputed_data[columns]=encoders_store[columns].inverse_transform(np.array(imputed_data[columns]).reshape(-1,1))
    return imputed_data

#CLEANING FILE
def clean_dataframe(df, impute = False, text_fields = [], date_fields = [], numeric_fields = [], categorical_fields = []):
    clean_df = (
      df
      #make the column names lower case and remove spaces
      .clean_names()

      #remove empty columns
      .remove_empty()

      #remove empty rows and columns
      .dropna(how='all')
    )

    #remove harmful characters. remove personal identifiers. make lowercase
    for field in text_fields:
        field = '_'.join(field.split()).lower()
        clean_df[field] = clean_df[field].fillna(' ').apply(fix_text)
        clean_df[field] = clean_df[field].apply(scrubadub.clean, replace_with='identifier')
        clean_df[field] = clean_df[field].str.lower()

    #impute missing values
    if impute:
        clean_df = impute_df(clean_df)

    #standardize the format of all date fields
    for field in date_fields:
        field = '_'.join(field.split()).lower()
        try:
          clean_df[field] = clean_df[field].apply(arrow.get)
        except ParserError:
          try:
            clean_df[field] = clean_df[field].apply(lambda x: arrow.get(datetime.strptime(x,'%m/%d/%Y')))
          except ValueError:
            clean_df[field] = clean_df[field].apply(lambda x: arrow.get(datetime.strptime(x,'%m/%d/%y')))

    #make sure all numeric fields have the proper data type
    for field in numeric_fields:
        field = '_'.join(field.split()).lower()
        clean_df[field] = pd.to_numeric(clean_df[field])

    #make sure all categorical variables have the proper data type
    for field in categorical_fields:
        field = '_'.join(field.split()).lower()
        clean_df[field] = clean_df[field].astype('category')

    clean_df=clean_df.clean_names()

    globals_file.clean_data_used = True

    return(clean_df)



from pymongo import MongoClient #MongoDB
import pandas as pd
import json
import datetime
import re
from dxc.ai.global_variables import globals_file
from dxc.ai.logging import pipeline_logging

def convert_dates_from_arrow_to_string(df, arrow_date_fields):
    for field in arrow_date_fields:
        df[field] = df[field].apply(format)
    return(df)

def write_raw_data(data_layer, raw_data, arrow_date_fields = []):
    ##make the column names lower case and remove spaces    
    if globals_file.clean_data_used == True:
        raw_data = raw_data.clean_names()
        globals_file.wrt_raw_data_used = True
        globals_file.clean_data_used = False    
    ##convert your raw data into writable data by converting Arrow dates to strings
    writable_raw_data = convert_dates_from_arrow_to_string(raw_data, arrow_date_fields)
    
    #connect to the data layer
    db_connect = MongoClient(data_layer["connection_string"])
    database=db_connect[data_layer['database_name']]
    collections = database.list_collection_names(include_system_collections=False)
    new_coll=database['meta_data']
    if 'meta_data' not in collections:

        meta_information = { "collection_name": data_layer["collection_name"], "inserted_date_time": datetime.datetime.now(),
                            'data_source':data_layer["data_source"], 'cleaner':data_layer["cleaner"]}
        new_coll.insert_one(meta_information)
        database[data_layer["collection_name"]].insert_many(writable_raw_data.to_dict('records'))
        database["%s_1"%data_layer["collection_name"]].insert_many(writable_raw_data.to_dict('records'))
        meta_information = { "collection_name": "%s_1"%data_layer["collection_name"],  "inserted_date_time": datetime.datetime.now(),
                            'data_source':data_layer["data_source"], 'cleaner':data_layer["cleaner"] }
        new_coll.insert_one(meta_information)

    else:
        if data_layer["collection_name"] in collections:
            tmp = data_layer["collection_name"]
            regex = re.compile(r'%s_\d+'%tmp)
            selected_files = list(filter(regex.search, collections))
            print(selected_files)
            # The list call is only required in Python 3, since filter was changed to return a generator
            length=len(selected_files)
            new_collection="%s_%d"%(data_layer["collection_name"],length+1)
            database[new_collection].insert_many(writable_raw_data.to_dict('records'))
            meta_information = { "collection_name": new_collection,  "inserted_date_time": datetime.datetime.now(),
                            'data_source':data_layer["data_source"], 'cleaner':data_layer["cleaner"] }
            new_coll.insert_one(meta_information)
            database.data_layer["collection_name"].drop()
            new_coll.delete_one({'collection_name':data_layer["collection_name"]})
            database[data_layer["collection_name"]].insert_many(writable_raw_data.to_dict('records'))
            meta_information = { "collection_name": data_layer["collection_name"],  "inserted_date_time": datetime.datetime.now(),
                            'data_source':data_layer["data_source"], 'cleaner':data_layer["cleaner"] }
            new_coll.insert_one(meta_information)
    return database[data_layer["collection_name"]]

#Handle case-sensitive for column names in pipeline
def col_header_conv_1(pipe):
    for key,value in pipe.items():
        if type(value) is dict:
            col_header_conv_1(value)
        elif type(value) is list:
            for each in value:
                if type(each) is dict:
                    col_header_conv_1(each)
                else:
                    if isinstance(each, str):
                        j = value.index(each)
                        value[j] = '_'.join(each.split()).lower()
        else:
            if isinstance(value, str):
                pipe[key] = '_'.join(value.split()).lower()
    return pipe

def col_header_conv(pipe):
    for i in range(len(pipe)):
        single_pipe = pipe[i]
        new_value = col_header_conv_1(single_pipe)
        pipe[i] = new_value
    return pipe

def access_data_from_pipeline(db, pipe):
    pipeline_logging.pipeline_log(pipe)
    if globals_file.wrt_raw_data_used == True or globals_file.clean_data_used == True:
        pipe = col_header_conv(pipe)
        globals_file.wrt_raw_data_used = False
        globals_file.clean_data_used = False

    data = db.aggregate(pipeline=pipe)
    data = list(data)
    df = pd.io.json.json_normalize(data)
    globals_file.run_experiment_warm_start = False
    return df



"""
Importing all functions into ai namespace
"""

from .AI_guild.AI_guild import guild_member_should_apply_for_badge
from .AI_guild.AI_guild import apply_for_an_ai_badge
from .AI_guild.AI_guild import AI_Guild_Role
from .AI_guild.AI_guild import AI_Badge

from .read_data.read_json import flatten_json_into_dataframe
from .read_data.read_json import get_file_path_json
from .read_data.read_json import read_data_frame_from_remote_json
from .read_data.read_json import read_data_frame_from_local_json
from .read_data.read_json import read_data_frame_from_local_json
from .read_data.read_excel import get_file_path_excel
from .read_data.read_excel import read_data_frame_from_local_excel_file
from .read_data.read_csv import get_file_path_csv
from .read_data.read_csv import read_data_frame_from_local_csv
from .read_data.read_csv import read_data_frame_from_remote_csv
from .read_data.read_json import read_data_frame_from_remote_json

from .clean_data.clean_data import encode
from .clean_data.clean_data import impute_df
from .clean_data.clean_data import clean_dataframe


from .visualization.visualization import explore_features
from .visualization.visualization import visualize_missing_data
from .visualization.visualization import plot_distributions
from .visualization.visualization import explore_complete_data

from .pipeline.pipeline import convert_dates_from_arrow_to_string
from .pipeline.pipeline import write_raw_data
from .pipeline.pipeline import access_data_from_pipeline

from .run_model.run_model import run_experiment
from .run_model.run_model import model
from .run_model.run_model import prediction
from .run_model.run_model import regression
from .run_model.run_model import classification
from .run_model.run_model import tpot_regression
from .run_model.run_model import tpot_classification
from .run_model.interpret_model import Global_Model_Explanation
from .run_model.interpret_model import Explanation_Dashboard
from .run_model.clustering import Clustering

from .publish_microservice.publish_microservice import publish_microservice

from .deep_learning.image_classifier import create_training_data
from .deep_learning.image_classifier import seggregate_data
from .deep_learning.image_classifier import split_normalize_data
from .deep_learning.image_classifier import image_classifier

from .logging.pipeline_logging          import pipeline_log
from .logging.experiment_design_logging import experiment_design_log
from .logging.microservice_logging      import microservice_design_log

from .datasets._base import load_data
from .datasets._base import load_data_details
# from .datasets._base import get_data




import numpy as np
import pandas as pd
import operator
import plotly
import sys
import datetime
from collections import defaultdict, Counter
from matplotlib import pyplot as plt
#%matplotlib inline

from statsmodels.tsa.stattools import adfuller
from statsmodels.tsa.seasonal import seasonal_decompose
from statsmodels.graphics.tsaplots import plot_acf, plot_pacf
from sklearn.metrics import mean_squared_error
from statsmodels.tsa.ar_model import AR
from statsmodels.tsa.arima_model import ARMA, ARIMA
from statsmodels.tsa.holtwinters import SimpleExpSmoothing, Holt, ExponentialSmoothing

import pyaf.ForecastEngine as autof

from pmdarima.arima import auto_arima
from pmdarima.model_selection import train_test_split

import warnings
warnings.filterwarnings('ignore', 'statsmodels.tsa.ar_model.AR', FutureWarning)


#Dickey-Fuller Test for stationary check
def adf_test(timeseries):
    #Perform Dickey-Fuller test:
    print('Results of Augmented Dickey-Fuller(ADF) Statistical Test:')
    dftest = adfuller(timeseries, autolag='AIC')
    dfoutput = pd.Series(dftest[0:4], index=['Test Statistic','p-value','#Lags Used','Number of Observations Used'])
    for key,value in dftest[4].items():
       dfoutput['Critical Value (%s)'%key] = value
    print(dfoutput)

#autoRegressiveModel
def autoRegressiveModel(df, no_predictions = 7, debug = False , visualize = False):
    data = df.values
    # Splitting data into train and test set.
    train, test = data[1:len(data)-no_predictions], data[len(data)-no_predictions:]
    # train autoregression
    model = AR(train)
    model_fit = model.fit()
    if(debug):
        print(model_fit.summary())
    # make predictions
    predictions = model_fit.predict(start=len(train), end=len(train)+len(test)-1+no_predictions, dynamic=False)
    if(visualize):
        plt.plot(test , color = "blue" , label = "testing data")
        plt.plot(predictions, color='red' , label = "prediction")
        plt.legend(loc='best')
        plt.show()
    error = np.sqrt(mean_squared_error(test, predictions[:no_predictions]))
    return (predictions[-no_predictions:] , error, model_fit)

def simpleExponentialSmoothing(df, no_predictions=7, debug = False , visualize = False):
    train_data, test_data = df[1:int(len(df)-no_predictions)], df[int(len(df)-no_predictions):]
    fit1 = SimpleExpSmoothing(np.asarray(train_data)).fit(smoothing_level=0.85 , optimized=False)
    if(debug):
         print(fit1.summary())
    predictions = fit1.forecast(no_predictions*2)
    if(visualize):
        plt.plot(list(test_data), color = 'blue', label='testing data')
        plt.plot(list(predictions), color = 'red',label='prediction')
        plt.legend(loc='upper left', fontsize=8)
        plt.show()
    error = np.sqrt(mean_squared_error(test_data, predictions[:no_predictions]))
    return (predictions[-no_predictions:] , error, fit1)

def doubleSmoothing(df, no_predictions=7, debug = False , visualize = False):
    train_data, test_data = df[1:int(len(df)-no_predictions)], df[int(len(df)-no_predictions):]
    model = ExponentialSmoothing(np.asarray(train_data), trend='add', seasonal=None)
    fit1 = model.fit()
    if(debug):
        print(fit1.summary())
    predictions = fit1.forecast(no_predictions*2)
    if(visualize):
        plt.plot(list(test_data), color = 'blue', label='testing data')
        plt.plot(list(predictions), color = 'red',label='prediction')
        plt.legend(loc='upper left', fontsize=8)
        plt.show()
    error = np.sqrt(mean_squared_error(test_data, predictions[:no_predictions]))
    return (predictions[-no_predictions:] , error, fit1)


def autoRegressiveMovingAverageModel(df, order=(1,0), no_predictions=7, debug=False , visualize=False):
    data = df.values
    train, test = data[1:len(data)-no_predictions], data[len(data)-no_predictions:]
    model = ARMA(train, order = order)
    model_fit = model.fit()
    if(debug):
        print(model_fit.summary())
    predictions = model_fit.predict(start=len(train), end=len(train)+len(test)-1+no_predictions, dynamic=False)
    if(visualize):
        plt.plot(test , color = "blue" , label = "testing data")
        plt.plot(predictions, color='red' , label = "prediction")
        plt.legend(loc='best')
        plt.show()
    error = np.sqrt(mean_squared_error(test, predictions[:no_predictions]))
    return (predictions[-no_predictions:] , error, model_fit)

def getBestForcastingModel(df, no_predictions=7, debug=False, visualize = False):
    modelResults = {}
    modelResults["autoRegressiveModel"] = autoRegressiveModel(df, no_predictions)[1]
    modelResults["simpleExponentialSmoothing"] = simpleExponentialSmoothing(df, no_predictions)[1]
    modelResults["doubleSmoothing"] = doubleSmoothing(df, no_predictions)[1]
    modelResults["autoRegressiveMovingAverageModel"] = autoRegressiveMovingAverageModel(df, order=(1,0), no_predictions=no_predictions)[1]
    bestModel = min(modelResults.items(), key=operator.itemgetter(1))[0]
    print(bestModel)
    print(adf_test(df))
    if bestModel == 'autoRegressiveModel':
        results, error, model = autoRegressiveModel(df, no_predictions, debug, visualize)
        print('RMSE:', error)
        return model
    elif bestModel == 'simpleExponentialSmoothing':
        results, error, model =  simpleExponentialSmoothing(df, no_predictions, debug, visualize)
        print('RMSE:', error)
        return model
    elif bestModel == 'doubleSmoothing':
        results, error, model = doubleSmoothing(df, no_predictions, debug, visualize)
        print('RMSE:', error)
        return model
    elif bestModel == 'autoRegressiveMovingAverageModel':
        results, error, model = autoRegressiveMovingAverageModel(df, order=(1,0), no_predictions=no_predictions, debug=debug, visualize=visualize)
        print('RMSE:', error)
        return model



#VISUALIZATION
from yellowbrick.features import Rank2D #exploring raw data
import matplotlib.pyplot as plt
import missingno as msno #gauge dataset completeness
import seaborn as sns #data exploration, distribution plotting
import pandas as pd
from datacleaner import autoclean
import math
from pandas.api.types import is_numeric_dtype
from pandas_profiling import ProfileReport


#VISUALIZATION

#display the correlations in pairwise comparisons of all features
def explore_features(df):
    df_copy = df.copy()

    #for some reason, the visualize doesn't accept categorical
    #variables. those have to be converted to strings
    for (col,data) in df_copy.iteritems():
        if df_copy[col].dtype.name == "category":
            df_copy[col] = df_copy[col].astype(str)

    numeric_df = autoclean(df_copy)
    visualizer = Rank2D(algorithm="pearson")
    visualizer.fit_transform(numeric_df)
    visualizer.poof()

#display a visual representation of missing fields in the given data
def visualize_missing_data(df):
    msno.matrix(df, figsize=(15,8))
    
def explore_complete_data(df, title='Complete Data Report'):
    profile = ProfileReport(df, title, html={'style':{'full_width':False}})
    return profile

#plot the distribution of values of each field in the given data
def plot_distributions(df):

    #set plot style
    sns.set(style="darkgrid")

    features = len(df.columns)

    #determine the number of columns in the plot grid and the width and height of each plot
    grid_cols = 3
    plot_width = 5
    plot_height = 3

    #determine the width of the plot grid and number of rows
    grid_width = plot_width * grid_cols
    num_rows = math.ceil(features/grid_cols)

    #determine the width of the plot grid
    grid_height = plot_height * num_rows

    #lay out the plot grid
    fig1 = plt.figure(constrained_layout=True, figsize = (grid_width,grid_height))
    gs = fig1.add_gridspec(ncols = grid_cols, nrows = num_rows)

    #step through the dataframe and add plots for each feature
    current_column = 0
    current_row = 0
    for col in df.columns:

        #set up a plot
        f1_ax1 = fig1.add_subplot(gs[current_row, current_column])
        f1_ax1.set_title(col)

        #create a plot for numeric values
        if is_numeric_dtype(df[col]):
            sns.distplot(df[col], ax = f1_ax1).set_xlabel('')
    
        #creare a plot for categorical values
        if df[col].dtype.name == "category":
            sns.countplot(df[col], ax = f1_ax1, order = df[col].value_counts().index).set_xlabel('')

        #move to the next column
        current_column +=1

        #determine if it is time to start a new row
        if current_column == grid_cols:
            current_column = 0
            current_row +=1



clean_data_used = False
wrt_raw_data_used = False
run_experiment_used = False
run_experiment_encoder = None
run_experiment_target_encoder = None
run_experiment_target_encoder_used = False
run_experiment_encoder_used = False
#below global variables are related to warm_start feature of auto_ml. Please don't change them. 
run_experiment_warm_start = False
run_experiment_x_train = None
run_experiment_x_test = None
run_experiment_y_train = None
run_experiment_y_test = None
run_experiment_model = None
run_experiment_datatransformed = None
run_experiment_labeldata = None


import json
import pandas as pd
import urllib.parse #input data
from flatten_json import flatten
from tkinter import Tk
from tkinter import filedialog
from enum import Enum

def flatten_json_into_dataframe(json_data):
    #flatten the nested JSON data into a data frame
    try:
        json_data_flattened = [flatten(d) for d in json_data]
        df = pd.DataFrame(json_data_flattened)
        return(df)
    except:
        try:
            df = pd.DataFrame(json_data).T
            return(df)
        except:
            raise Exception("Uploaded JSON file is not in proper structure. Please choose different file.")

#Read JSON file from remote
def read_data_frame_from_remote_json(json_url):
    with urllib.request.urlopen(json_url) as url:
        json_data = json.loads(url.read().decode())
    df = flatten_json_into_dataframe(json_data)
    return(df)

def get_file_path_json():
    root = Tk()
    root.update()
    def open_file():
        file = filedialog.askopenfilename(filetypes=[("Json files", "*.json")]) 
        return file
    file_path = open_file()
    root.destroy()
    return file_path


def read_data_frame_from_local_json():
    try:
        from google.colab import files
        IN_COLAB = True
    except:
        IN_COLAB = False
    if IN_COLAB:
        
        uploaded = files.upload()
        file_name = list(uploaded.keys())[0]
        json_data = json.loads(uploaded[file_name])
        df = flatten_json_into_dataframe(json_data)
        return df
    else:
        json_path = get_file_path_json()
        json_data = json.load(open(json_path))
        df = flatten_json_into_dataframe(json_data)
        return df

def read_data_frame_from_remote_json(json_url):
    with urllib.request.urlopen(json_url) as url:
        json_data = json.loads(url.read().decode())
    df = flatten_json_into_dataframe(json_data)
    return(df)


import Algorithmia
from Algorithmia.errors import AlgorithmException
import shutil #serializing models
import urllib.parse #input data
from git import Git, Repo, remote
import os
import pickle
from IPython.display import YouTubeVideo
from IPython.core.magic import register_line_cell_magic
import urllib.request, json
from dxc.ai.global_variables import globals_file
from dxc.ai.logging import microservice_logging

def publish_microservice(microservice_design, trained_model, verbose = False):
    #Capture microservice_design in log
    microservice_logging.microservice_design_log(microservice_design)
    # create a connection to algorithmia
    client=Algorithmia.client(microservice_design["api_key"])
    api = client.algo(microservice_design["execution_environment_username"] + "/" + microservice_design["microservice_name"])

    # create the algorithm if it doesn't exist
    try:
        api.create(
            details = {
                "label": microservice_design["microservice_name"],
            },
            settings = {
                "language": "python3-1",
                "source_visibility": "closed",
                "license": "apl",
                "network_access": "full",
                "pipeline_enabled": True,
                "environment": "cpu"
            }
    )
    except Exception as error:
            print(error)

    # create data collection if it doesn't exist
    if not client.dir(microservice_design["model_path"]).exists():
        client.dir(microservice_design["model_path"]).create()

    # define a local work directory
    local_dir = microservice_design["microservice_name"]

    # delete local directory if it already exists
    if os.path.exists(local_dir):
        shutil.rmtree(local_dir)

    # create local work directory
    os.makedirs(local_dir)

    # serialize the model locally
    local_model = "{}/{}".format(local_dir, "mdl")

    # open a file in a specified location
    file = open(local_model, 'wb')
    # dump information to that file
    pickle.dump(trained_model, file)
    # close the file
    file.close()

    # upload our model file to our data collection
    api_model = "{}/{}".format(microservice_design["model_path"], microservice_design["microservice_name"])
    client.file(api_model).putFile(local_model)

    if globals_file.run_experiment_encoder_used:
        encode_model = 'encode_file.pkl'
        encode_output = open(encode_model, 'wb')
        pickle.dump(globals_file.run_experiment_encoder, encode_output)
        encode_output.close()
        encode_folder =  microservice_design["microservice_name"] + '_encoder'
        encode_path = "{}/{}".format(microservice_design["model_path"], encode_folder)
        client.file(encode_path).putFile(encode_model)
    if globals_file.run_experiment_target_encoder_used:
        target_encode_model = 'target_encode_file.pkl'
        target_encode_output = open(target_encode_model, 'wb')
        pickle.dump(globals_file.run_experiment_target_encoder, target_encode_output)
        target_encode_output.close()
        target_encode_folder =  microservice_design["microservice_name"] + '_target_encoder'
        target_encode_path = "{}/{}".format(microservice_design["model_path"], target_encode_folder)
        client.file(target_encode_path).putFile(target_encode_model)

    # encode API key, so we can use it in the git URL
    encoded_api_key = urllib.parse.quote_plus(microservice_design["api_key"])

    algo_repo = "https://{}:{}@git.algorithmia.com/git/{}/{}.git".format(
        microservice_design["execution_environment_username"],
        encoded_api_key,
        microservice_design["execution_environment_username"],
        microservice_design["microservice_name"]
        )

    class Progress(remote.RemoteProgress):
        if verbose == False:
            def line_dropped(self, line):
                pass
            def update(self, *args):
                pass
        else:
            def line_dropped(self, line):
                print(line)
            def update(self, *args):
                print(self._cur_line)

    p = Progress()

    try:
        Repo.clone_from(algo_repo, "{}/{}".format(local_dir, microservice_design["microservice_name"]), progress=p)
        cloned_repo = Repo("{}/{}".format(local_dir, microservice_design["microservice_name"]))
    except Exception as error:
        print("here")
        print(error)

    api_script_path = "{}/{}/src/{}.py".format(local_dir, microservice_design["microservice_name"], microservice_design["microservice_name"])
    dependency_file_path = "{}/{}/{}".format(local_dir, microservice_design["microservice_name"], "requirements.txt")

    # defines the source for the microservice
    results = "{'results':prediction}"
    file_path = "'" + api_model + "'"
    if globals_file.run_experiment_encoder_used:
        encodefile_path = "'" + encode_path + "'"
    if globals_file.run_experiment_target_encoder_used:
        target_encodefile_path = "'" + target_encode_path + "'"
    ##Don't change the structure of below docstring
    ##this is the source code needed for the microservice
    
    src_code_content = """import Algorithmia
import auto_ml
import pandas as pd
import pickle
# create an Algorithmia client
client = Algorithmia.client()
def load_model():
    # Get file by name
    # Open file and load model
\tfile_path = {file_path}
\tmodel_path = client.file(file_path).getFile().name
    # Open file and load model
\twith open(model_path, 'rb') as f:
\t\tmodel = pickle.load(f)
\t\treturn model
trained_model = load_model()
def apply(input):
\tprediction = trained_model.predict(input)
\treturn {results}"""

    ## source code for customized model
    src_code_generalized = """import Algorithmia
import auto_ml
import pandas as pd
import pickle
import json
import numpy as np
# create an Algorithmia client
client = Algorithmia.client()
def load_model():
    # Get file by name
    # Open file and load model
\tfile_path = {file_path}
\tmodel_path = client.file(file_path).getFile().name
    # Open file and load model
\twith open(model_path, 'rb') as f:
\t\tmodel = pickle.load(f)
\t\treturn model
trained_model = load_model()
def default(obj):
    if type(obj).__module__ == np.__name__:
        if isinstance(obj, np.ndarray):
            return obj.tolist()
        else:
            return obj.item()
    raise TypeError('Unknown type:', type(obj))
def apply(input):
\tprediction = trained_model.predict(input)
\tprediction = json.dumps(prediction, default=default)
\treturn {results}"""

    ## source code for generalized tpot model
    src_code_generalized_encode = """import Algorithmia
import auto_ml
import pandas as pd
import pickle
import json
import numpy as np
import feature_engine
# create an Algorithmia client
client = Algorithmia.client()
def load_model():
    # Get file by name
    # Open file and load model
\tfile_path = {file_path}
\tmodel_path = client.file(file_path).getFile().name
    # Open file and load model
\twith open(model_path, 'rb') as f:
\t\tmodel = pickle.load(f)
\t\treturn model
trained_model = load_model()
def load_encode():
    # Get file by name
    # Open file and load encoder
\tencodefile_path = {encodefile_path}
\tencode_path = client.file(encodefile_path).getFile().name
    # Open file and load encoder
\twith open(encode_path, 'rb') as f:
\t\tencoder = pickle.load(f)
\t\treturn encoder
encode = load_encode()
def default(obj):
    if type(obj).__module__ == np.__name__:
        if isinstance(obj, np.ndarray):
            return obj.tolist()
        else:
            return obj.item()
    raise TypeError('Unknown type:', type(obj))
def apply(input):
\tinput = pd.DataFrame([input])
\ttry:
\t\tinput = encode.transform(input)
\texcept:
\t\tpass
\tprediction = trained_model.predict(input)
\tprediction = json.dumps(prediction[0], default=default)
\treturn {results}"""

    ## source code for generalized tpot model
    src_code_generalized_target_encode = """import Algorithmia
import auto_ml
import pandas as pd
import pickle
import json
import numpy as np
import feature_engine
# create an Algorithmia client
client = Algorithmia.client()
def load_model():
    # Get file by name
    # Open file and load model
\tfile_path = {file_path}
\tmodel_path = client.file(file_path).getFile().name
    # Open file and load model
\twith open(model_path, 'rb') as f:
\t\tmodel = pickle.load(f)
\t\treturn model
trained_model = load_model()
def load_encode():
    # Get file by name
    # Open file and load encoder
\tencodefile_path = {encodefile_path}
\tencode_path = client.file(encodefile_path).getFile().name
    # Open file and load encoder
\twith open(encode_path, 'rb') as f:
\t\tencoder = pickle.load(f)
\t\treturn encoder
encode = load_encode()
def load_target_encode():
    # Get file by name
    # Open file and load target encoder
\ttarget_encodefile_path = {target_encodefile_path}
\ttarget_encode_path = client.file(target_encodefile_path).getFile().name
    # Open file and load target encoder
\twith open(target_encode_path, 'rb') as f:
\t\ttarget_encoder = pickle.load(f)
\t\treturn target_encoder
target_encode = load_target_encode()

def default(obj):
    if type(obj).__module__ == np.__name__:
        if isinstance(obj, np.ndarray):
            return obj.tolist()
        else:
            return obj.item()
    raise TypeError('Unknown type:', type(obj))
def apply(input):
\tinput = pd.DataFrame([input])
\ttry:
\t\tinput = encode.transform(input)
\texcept:
\t\tpass
\tprediction = trained_model.predict(input)
\ttry:
\t\tprediction = target_encode.inverse_transform(prediction)
\t\tprediction = prediction[0]
\texcept:
\t\tprediction = json.dumps(prediction[0], default=default)
\treturn {results}"""

    if globals_file.run_experiment_used:
        src_code_content = src_code_generalized
        if globals_file.run_experiment_encoder_used:
            src_code_content = src_code_generalized_encode
        if globals_file.run_experiment_target_encoder_used:
            src_code_content = src_code_generalized_target_encode       

    splitted=src_code_content.split('\n')

    ##writes the source into the local, cloned GitHub repository
    with open(api_script_path, "w") as f:
        for line in splitted:
            if line.strip()=="file_path = {file_path}":
                line="\tfile_path = {}".format(file_path)
            if line.strip()=="encodefile_path = {encodefile_path}":
                line="\tencodefile_path = {}".format(encodefile_path)
            if line.strip()=="target_encodefile_path = {target_encodefile_path}":
                line="\ttarget_encodefile_path = {}".format(target_encodefile_path)
            if line.strip()=="return {results}":
                line="\treturn {}".format(results)
            f.write(line + '\n')

    ##Don't change the structure of below docstring
    ##this is the requirements needed for microservice
    requirements_file_content="""algorithmia>=1.0.0,<2.0
    six
    auto_ml
    pandas
    numpy
    feature-engine
    bottleneck==1.2.1"""

    post_split=requirements_file_content.split('\n')

    #writes the requirements file into the local, cloned GitHub repository.
    with open(dependency_file_path, "w") as f:
        for line in post_split:
            line = line.lstrip()
            f.write(line + '\n')

    # Publish the microservice

    files = ["src/{}.py".format(microservice_design["microservice_name"]), "requirements.txt"]
    cloned_repo.index.add(files)
    cloned_repo.index.commit("Add algorithm files")
    origin = cloned_repo.remote(name='origin')

    p = Progress()

    origin.push(progress=p)

    # publish/deploy our algorithm
    client.algo(microservice_design["api_namespace"]).publish()

    #  code generates the api endpoint for the newly published microservice
    latest_version = client.algo(microservice_design["api_namespace"]).info().version_info.semantic_version
    api_url = "https://api.algorithmia.com/v1/algo/{}/{}".format(microservice_design["api_namespace"], latest_version)

    return api_url



import json
import requests
from enum import Enum

class AI_Guild_Role(Enum):
    PROJECT_MANAGER = 1,
    DATA_SCIENTIST = 2,
    DATA_ENGINEER = 3,
    ALL = 4

class AI_Badge(Enum):
    CREATE_DATA_STORIES = 1
    RUN_AGILE_TRANSFORMATION = 2
    BUILD_DATA_PIPELINES = 3
    RUN_AI_EXPERIMENT = 4
    BUILD_UTILITY_AI_SERVICES = 5
    PERFORM_AI_FORENSICS = 6
    TEST = 7


def guild_member_should_apply_for_badge(guild_member_roles, ai_badge):


    guild_role_badges = {
        AI_Guild_Role.PROJECT_MANAGER : [AI_Badge.CREATE_DATA_STORIES, AI_Badge.RUN_AGILE_TRANSFORMATION],
        AI_Guild_Role.DATA_SCIENTIST : [AI_Badge.RUN_AI_EXPERIMENT, AI_Badge.PERFORM_AI_FORENSICS],
        AI_Guild_Role.DATA_ENGINEER : [AI_Badge.BUILD_DATA_PIPELINES, AI_Badge.BUILD_UTILITY_AI_SERVICES],
        AI_Guild_Role.ALL: [AI_Badge.CREATE_DATA_STORIES, AI_Badge.RUN_AGILE_TRANSFORMATION, AI_Badge.BUILD_DATA_PIPELINES, AI_Badge.RUN_AI_EXPERIMENT,            AI_Badge.BUILD_UTILITY_AI_SERVICES, AI_Badge.PERFORM_AI_FORENSICS]
        }

    ai_badge_id = {
        AI_Badge.CREATE_DATA_STORIES: "dd05bbdf-ad5b-469d-ab2c-4dd218fd68fe",
        AI_Badge.RUN_AGILE_TRANSFORMATION: "8ec48861-c355-4e44-b98f-0a80cf1440a8",
        AI_Badge.BUILD_DATA_PIPELINES: "2cffc101-8fc3-4680-a8cd-29ec58483832",
        AI_Badge.RUN_AI_EXPERIMENT: "ddeb2020-1db5-48c6-8b2c-ea2e50f050d7",
        AI_Badge.BUILD_UTILITY_AI_SERVICES: "6e8b661f-31bc-46f7-89e4-194e7e6ebb21",
        AI_Badge.PERFORM_AI_FORENSICS: "ec96e016-9e33-4b4e-a6bd-43491e811179",
        AI_Badge.TEST: "b828e318-8501-434e-9f55-ccdb7000ee09"
    }
    #start by assuming we should not apply for the badge
    apply_for_badge = False
    for role in guild_member_roles:
        #apply for the badge if a matching role is found
        if ai_badge in guild_role_badges[role]: apply_for_badge = True
    return apply_for_badge

def apply_for_an_ai_badge(ai_guild_profile, ai_badge):

    guild_role_badges = {
        AI_Guild_Role.PROJECT_MANAGER : [AI_Badge.CREATE_DATA_STORIES, AI_Badge.RUN_AGILE_TRANSFORMATION],
        AI_Guild_Role.DATA_SCIENTIST : [AI_Badge.RUN_AI_EXPERIMENT, AI_Badge.PERFORM_AI_FORENSICS],
        AI_Guild_Role.DATA_ENGINEER : [AI_Badge.BUILD_DATA_PIPELINES, AI_Badge.BUILD_UTILITY_AI_SERVICES],
        AI_Guild_Role.ALL: [AI_Badge.CREATE_DATA_STORIES, AI_Badge.RUN_AGILE_TRANSFORMATION, AI_Badge.BUILD_DATA_PIPELINES, AI_Badge.RUN_AI_EXPERIMENT,            AI_Badge.BUILD_UTILITY_AI_SERVICES, AI_Badge.PERFORM_AI_FORENSICS]
    }

    ai_badge_id = {
        AI_Badge.CREATE_DATA_STORIES: "dd05bbdf-ad5b-469d-ab2c-4dd218fd68fe",
        AI_Badge.RUN_AGILE_TRANSFORMATION: "8ec48861-c355-4e44-b98f-0a80cf1440a8",
        AI_Badge.BUILD_DATA_PIPELINES: "2cffc101-8fc3-4680-a8cd-29ec58483832",
        AI_Badge.RUN_AI_EXPERIMENT: "ddeb2020-1db5-48c6-8b2c-ea2e50f050d7",
        AI_Badge.BUILD_UTILITY_AI_SERVICES: "6e8b661f-31bc-46f7-89e4-194e7e6ebb21",
        AI_Badge.PERFORM_AI_FORENSICS: "ec96e016-9e33-4b4e-a6bd-43491e811179",
        AI_Badge.TEST: "b828e318-8501-434e-9f55-ccdb7000ee09"
    }
    
    # Construct apiEndponit string
    apiPath = f'badges/{ai_badge_id[ai_badge]}/assertions'
    apiEndpoint = f'{ai_guild_profile["badge_platform_apiHost"]}{ai_guild_profile["badge_platform_apiBasePath"]}{apiPath}'  

    headers = {
        'Content-Type': 'application/json',
        'X-Api-Key': ai_guild_profile["badge_platform_apiKey"]
    }

    #for each member in the guild:
    num_guild_members = len(ai_guild_profile['guild_members'])
    for i in range(1,num_guild_members + 1):
    #apply for the badge if applicable
        if guild_member_should_apply_for_badge(ai_guild_profile['guild_members'][i]['roles'], ai_badge):
            payload = {
            'email': ai_guild_profile['guild_members'][i]['badge_applicant_email'],
            'evidence': ai_guild_profile["badge_evidence"]
            }
            response = requests.post(
                apiEndpoint,
                headers=headers,
                json=payload
            )
  
            print(response)
            print(json.dumps(response.json(), indent=4))



from interpret_community import TabularExplainer
from interpret_community.widget import ExplanationDashboard
import warnings

def Global_Model_Explanation(model,x_train,x_test,feature_names = None,classes = None, explantion_data = None):
    warnings.filterwarnings('ignore')
    #Using SHAP TabularExplainer
    explainer = TabularExplainer(model, 
                                  x_train, 
                                  features=feature_names, 
                                  classes=classes)
    #Generate global explanations
    if explantion_data == 'Training':
        global_explanation = explainer.explain_global(x_train)
    else:
        global_explanation = explainer.explain_global(x_test)
    ##print the global importance rank data
    print('global importance rank: {}'.format(global_explanation.get_feature_importance_dict()))
    #Return Generated Explanation dashboard
    return global_explanation

def Explanation_Dashboard(global_explanation, model, x_train, x_test, explantion_data = None):
    if explantion_data == 'Training':
        ExplanationDashboard(global_explanation, model, datasetX=x_train)
    else:
        ExplanationDashboard(global_explanation, model, datasetX=x_test)

    



from os.path import dirname, join
import csv
import pandas as pd

def load_filedata(module_path, data_file_name):
    df = pd.DataFrame()
    df1 = pd.DataFrame()
    try:
        with open(join(module_path, 'data', data_file_name)) as csv_file:
            data_file = csv.reader(csv_file)
            for i, ir in enumerate(data_file):
                if i == 0:
                    header = ir
                    df = pd.DataFrame(columns = header)
                else:
                    df1 = pd.DataFrame([ir],columns = header)
                df = df.append(df1, ignore_index = True)
        return df
    except:
        message = data_file_name[:-4] + " dataset is not available"
        df = pd.DataFrame([message])
        return df


##load data
def load_data(filename):
    module_path = dirname(__file__)
    finalfile = filename + '.csv'
    df = load_filedata(module_path, finalfile)
    return df

#load file content
def load_data_details(filename):
    module_path = dirname(__file__)
    finalfile = filename + '.txt'
    try:
        f = open(join(module_path, 'data', finalfile))
        file_content = f.read()
        return file_content
    except:
        message = "No Details found for " + filename + " data set"
        return message
    
# def get_data(dataset, display_details=False, save_copy=False):

#     """
#     >>> data = get_data('iris')
#     >>> data = get_data('iris', display_details=True)
#     >>> data = get_data('iris', display_details=True. save_copy=True)

#     """

#     import pandas as pd
#     import os.path
#     import requests
    
#     address = "https://raw.githubusercontent.com/dxc-technology/DXC-Industrialized-AI-Starter/master/dxc/ai/datasets/data/"
#     extension = ".csv"
#     filename = str(dataset) + extension

#     complete_address = address + filename

#     if os.path.isfile(filename):
#         data = pd.read_csv(filename)
#     else:
#         data = pd.read_csv(complete_address)

#     if save_copy:
#         save_name = filename
#         data.to_csv(save_name, index=False)
    
#     if display_details:
#         details_extension = ".txt"
#         details_filename = str(dataset) + details_extension
#         details_complete_address = address + details_filename
#         print(requests.get(details_complete_address).text)
#     return data






from setuptools import setup, find_packages

def readme():
    with open('README.md') as f:
        README = f.read()
    return README


setup(
    name="DXC-RL",
    version="1.0.3.2",
    description="Python library which is extensively used for all AI projects",
    long_description=readme(),
    long_description_content_type="text/markdown",
    url="https://github.com/SureshAthanti/DXC-Industrialized-AI-Starter",
    #download_url = "https://github.com/dxc-technology/DXC-Industrialized-AI-Starter/releases/tag/v1.0.tar.gz",
    author="DXC",
    license="Apache License 2.0",
    platforms='any',
    classifiers=[
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python :: 3.7",
    ],
    #packages=["dxc", "dxc.ai"],
    packages=find_packages(),
#     include_package_data=True,
    install_requires=["JIRA","scikit-learn==0.22.2.post1","auto_ml","Algorithmia","gitpython","flatten_json==0.1.7","pyjanitor","ftfy","arrow","pandas-profiling[notebook]==2.9.0",
                      "scrubadub","yellowbrick==1.1","datacleaner","missingno","pymongo","IPython","dnspython","pmdarima","pyaf","interpret-community","flask_cors","gevent","tpot",
                      "gym","keras-rl2","tensorflow"],
    entry_points={
        "console_scripts": [
            "dxc=dxc.ai:main",
        ]
    },
    package_data={'datasets/data': ['datasets/data/*'],},
    include_package_data=True,
)



import json
import pandas as pd
import urllib.parse #input data
from tkinter import Tk
from tkinter import filedialog
from enum import Enum
import io

def get_file_path_csv():
    root = Tk()
    root.update()
    def open_file():
        file = filedialog.askopenfilename(filetypes=[("csv files", "*.csv")])  
        return file
    file_path = open_file()
    root.destroy()
    return file_path



def read_data_frame_from_local_csv(col_names = [], sep=',', delim_whitespace=False, header = 'infer', 
                                   names=None, skiprows=None, error_bad_lines=True, encoding=None):
    try:
        from google.colab import files
        IN_COLAB = True
    except:
        IN_COLAB = False
    
    if IN_COLAB:
        
        uploaded = files.upload()
        csv_file_name = list(uploaded.keys())[0]
        df = pd.read_csv(io.BytesIO(uploaded[csv_file_name]), sep=sep, delim_whitespace=delim_whitespace, header = header,
                         names=names, skiprows=skiprows, error_bad_lines=error_bad_lines, encoding=encoding)
        if col_names != []:
            df.columns = col_names
        return(df)
    
    else:     
        csv_path = get_file_path_csv()  
        df = pd.read_csv(csv_path, delim_whitespace=delim_whitespace, header = header)
        if col_names != []:
            df.columns = col_names
        return(df)
        
def read_data_frame_from_remote_csv(csv_url, col_names = [], delim_whitespace=False, header = 'infer'):
    df = pd.read_csv(csv_url, delim_whitespace=delim_whitespace, header = header)
    if col_names != []:
        df.columns = col_names
    return(df)



import numpy as np
import gym
import rl

from tensorflow.keras.models import Sequential, Model
from tensorflow.keras.layers import Dense, Activation, Flatten, Input, Concatenate
from tensorflow.keras.optimizers import Adam

from rl.agents.dqn import DQNAgent
from rl.agents import DDPGAgent
from rl.policy import BoltzmannQPolicy
from rl.memory import SequentialMemory
from rl.random import OrnsteinUhlenbeckProcess

def define_policy(model_name):
  policy = []
  if (model_name == "DQN"):
    policy = BoltzmannQPolicy()
  return policy